import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.TextInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

public class PreProcessInputMatrix extends Configured implements Tool {
	
	private static String input;
    private static String output;
    
    private static int inputM;
	private static int inputN;
	
	private static int inputRP; // rows are split in to RP parts
	private static int inputCP; // columns are split in to CP parts
    
	public static class Map extends Mapper<LongWritable, Text, IntWritable, Text> {
	    
		private int M;
		private int N;
		
		private int RP; // rows are split in to RP parts
		private int CP; // columns are split in to CP parts
		
		private int RperPart;
		private int CperPart;
		
		private int startindex = 1; //first row or column indexed by 1
		
		@Override
		protected void setup(Context context) throws IOException,
				InterruptedException {
			
			M = Integer.parseInt(context.getConfiguration().get("M"));
			N = Integer.parseInt(context.getConfiguration().get("N"));
			RP = Integer.parseInt(context.getConfiguration().get("RP"));
			CP = Integer.parseInt(context.getConfiguration().get("CP"));
			
			if(M%RP == 0)
				RperPart = M/RP;
			else
				RperPart = M/RP + 1;
			
			if(N%CP == 0)
				CperPart = N/CP;
			else
				CperPart = N/CP + 1;
			
			System.out.println("RperPart: " + RperPart);
			System.out.println("CperPart: " + CperPart);
		}
		
	    public void map(LongWritable key, Text value, Context context) throws IOException, InterruptedException {
	    	
	    	String[] vals = value.toString().split(";");
			
			int row = 0;
			int column = 0;
			
			int highrow = 0;
			int highcolumn = 0;
			int newkey = 0;
			
			for(int i=0; i<vals.length; i++)
			{
				String[] fields = vals[i].split("\t");
				row = Integer.parseInt(fields[0]);
				column = Integer.parseInt(fields[1]);
				
				highrow = (row-startindex)/RperPart;
				highcolumn = (column-startindex)/CperPart;
				
				newkey = highrow*CP + highcolumn;
				
				context.write(new IntWritable(newkey), new Text((row-startindex) + "," + (column-startindex) + "," + fields[2]));
			}
	    }
	 } 
	        
	 public static class Reduce extends Reducer<IntWritable, Text, NullWritable, NullWritable> { 
		
		private int M;
		private int N;
			
		private int RP; // rows are split in to RP parts
		private int CP; // columns are split in to CP parts
			
		private int RperPart;
		private int CperPart;
		
		private int id;
		
		@Override
		protected void setup(Context context) throws IOException,
				InterruptedException {
			
			M = Integer.parseInt(context.getConfiguration().get("M"));
			N = Integer.parseInt(context.getConfiguration().get("N"));
			RP = Integer.parseInt(context.getConfiguration().get("RP"));
			CP = Integer.parseInt(context.getConfiguration().get("CP"));
			
			id = context.getConfiguration().getInt("mapred.task.partition", 0);
			
			if(M%RP == 0)
				RperPart = M/RP;
			else
				RperPart = M/RP + 1;
			
			if(N%CP == 0)
				CperPart = N/CP;
			else
				CperPart = N/CP + 1;
			
			System.out.println("RperPart: " + RperPart);
			System.out.println("CperPart: " + CperPart);
		}
			
	    public void reduce(IntWritable key, Iterable<Text> values, Context context) 
	      throws IOException, InterruptedException {
	    	
	    	FileSystem fs = FileSystem.get(context.getConfiguration());
	    	FSDataOutputStream out = null;
	    	BufferedWriter writer = null;
	    	
	    	boolean initialized = false;
			
	        for (Text val : values) {
	        	
	        	if(!initialized)
	        	{
	        		initialized = true;
	        		
	        		String[] fields = val.toString().split(",");
					int row = Integer.parseInt(fields[0]);
					int column = Integer.parseInt(fields[1]);
					
					int highrow = row/RperPart;
					int highcolumn = column/CperPart;
					
	        		String output = context.getConfiguration().get("outPath");
	        		//System.out.println("output path: " + output);
	    	    	out = fs.create(new Path(output + "/" + id + "-" + highrow + "-" + highcolumn));
	    	    	writer = new BufferedWriter(new OutputStreamWriter(out));
	        	}
	        	
	        	/*
	        	String[] fields = val.toString().split(",");
	        	
	        	if(fields.length == 3)
	        		writer.write(fields[0] + "," + fields[1] + "," + fields[2] + "\n");
	        	else
	        		System.out.println(val.toString());
	        	*/
	        	
	        	writer.write(val.toString() + "\n");
	        }
	        
	        writer.flush();
	        writer.close();
	        out.close();
	    }
	 }
	
	public static void preProcess() throws IOException,
	InterruptedException, ClassNotFoundException {
		//job1
		Configuration conf = new Configuration();
		
		conf.set("M", inputM + "");
		conf.set("N", inputN + "");
		conf.set("RP", inputRP + "");
		conf.set("CP", inputCP + "");
		conf.set("outPath", output);
		
		//avoid duplicated output files
		conf.setBoolean("mapred.map.tasks.speculative.execution", false);
		conf.setBoolean("mapred.reduce.tasks.speculative.execution", false);

		Job job1 = new Job(conf, "PreProcessInputMatrix");
		job1.setJarByClass(PreProcessInputMatrix.class);
		job1.setNumReduceTasks(100);
		job1.setJobName("PreProcessInputMatrix");
		
		job1.setMapOutputKeyClass(IntWritable.class);
		job1.setMapOutputValueClass(Text.class);
		job1.setOutputKeyClass(NullWritable.class);
		job1.setOutputValueClass(NullWritable.class);
	    
		job1.setMapperClass(Map.class);
		job1.setReducerClass(Reduce.class);
		job1.setInputFormatClass(TextInputFormat.class);
		job1.setOutputFormatClass(TextOutputFormat.class);
		
		
		FileInputFormat.addInputPath(job1, new Path(input));
		FileOutputFormat.setOutputPath(job1, new Path("empty"+output));
		job1.waitForCompletion(true);
	}

	public static void main(String[] args) throws Exception {

		int res = ToolRunner.run(new Configuration(), new PreProcessInputMatrix(), args);
        System.exit(res);
	}

	@Override
	public int run(String[] args) throws Exception {
		// TODO Auto-generated method stub
	   if (args.length != 6) {
             System.err.println("Usage: PreProcessInputMatrix <input> <output> <m> <n> <rp> <cp>");
             System.exit(2);
       }

       input = args[0];
       output = args[1];
       
       inputM  = Integer.parseInt(args[2]);
       inputN  = Integer.parseInt(args[3]);
       inputRP  = Integer.parseInt(args[4]);
       inputCP  = Integer.parseInt(args[5]);
       
       preProcess();
       
		return 0;
	}
	
}
