package frequentupdate;

import java.io.IOException;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.WritableComparable;
import org.apache.hadoop.io.WritableComparator;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Partitioner;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;

import util.NonSplittableTextInputFormat;

import auxstructure.IntPair;

public class MainNMFfre {
	
	public static class FirstPartitioner<IntPair, Text> extends Partitioner<IntPair, Text> {

		@Override
		public int getPartition(IntPair key, Text value, int numPartitions) {
			return   ((auxstructure.IntPair) key).getFirst() % numPartitions;
		}
	}
    
    
	public static class KeyComparator extends WritableComparator {
		
		protected KeyComparator() {
			super(IntPair.class, true);
		}

		@Override
		public int compare(WritableComparable w1, WritableComparable w2) {
			IntPair ip1 = (IntPair) w1;
			IntPair ip2 = (IntPair) w2;
			int cmp = ip1.getFirst() - ip2.getFirst();
			if (cmp != 0) {
				return cmp;
			}
			
			return ip1.getSecond() - ip2.getSecond();
		}
	}
	

	public static class GroupComparator extends WritableComparator {
		protected GroupComparator() {
			super(IntPair.class, true);
		}

		@Override
		public int compare(WritableComparable w1, WritableComparable w2) {
			IntPair ip1 = (IntPair) w1;
			IntPair ip2 = (IntPair) w2;
			return ip1.getFirst() - ip2.getFirst();
		}
	}
	
	public static void computeH(Path input, String middledir, int k, int itr, int m, int n, int rp, int cp) throws IOException,
	InterruptedException, ClassNotFoundException {
		//job1
		Configuration conf = new Configuration();
		String wpath = middledir + "/wmatrix/" + (itr-1);
		conf.set("middledir", middledir);
		//conf.setBoolean("mapred.output.compression.codec", true);
		//conf.setClass("mapred.output.compression.codec", GzipCodec.class,CompressionCodec.class);
		//conf.setInt("mapred.inmem.merge.threshold",0);
		//conf.setFloat("mapred.job.reduce.input.buffer.percent", 1);
		conf.set("mapred.child.java.opts", "-Xmx400m");
		//conf.setInt("io.sort.mb", 200);
		conf.setInt("io.sort.factor", 100);
		//conf.setInt("mapred.reduce.parallel.copies",10);
		conf.setInt("io.file.buffer.size", 128);
		conf.set("k", k + "");
		conf.set("itr", itr + "");
		conf.set("M", m + "");
		conf.set("N", n + "");
		conf.set("RP", rp + "");
		conf.set("CP", cp + "");
		Job job1 = new Job(conf, "computingWW"+itr);
		job1.setJarByClass(UpdateH.class);
		job1.setNumReduceTasks(1);
		job1.setJobName("computingWW"+itr);
		job1.setOutputKeyClass(IntWritable.class);
		job1.setOutputValueClass(Text.class);
		job1.setMapperClass(UpdateH.hmap1.class);
		job1.setReducerClass(UpdateH.hreduce1.class);
		job1.setInputFormatClass(NonSplittableTextInputFormat.class);
		job1.setOutputFormatClass(TextOutputFormat.class);


		FileInputFormat.addInputPath(job1, new Path(wpath));
		String outstr = middledir + "/ww/" + itr;
		FileOutputFormat.setOutputPath(job1, new Path(outstr));
		job1.waitForCompletion(true);


		//job2
		//avoid duplicated output files
		conf.setBoolean("mapred.map.tasks.speculative.execution", false);
		conf.setBoolean("mapred.reduce.tasks.speculative.execution", false);

		Job job2 = new Job(conf, "updateH"+itr);
		job2.setJarByClass(UpdateH.class);
		job2.setJobName("updateH"+itr);
		job2.setMapOutputKeyClass(IntPair.class);
		job2.setMapOutputValueClass(Text.class);
		job2.setOutputKeyClass(IntWritable.class);
		job2.setOutputValueClass(Text.class);
		job2.setNumReduceTasks(4);
		job2.setMapperClass(UpdateH.hmap2.class);
		job2.setReducerClass(UpdateH.hreduce2.class);
		//job2.setCombinerClass(hreduce2.class);
		job2.setInputFormatClass(NonSplittableTextInputFormat.class);
		job2.setOutputFormatClass(TextOutputFormat.class);

		//secondary sort
		job2.setPartitionerClass(FirstPartitioner.class);
		job2.setGroupingComparatorClass(GroupComparator.class);
		job2.setSortComparatorClass(KeyComparator.class);

		FileInputFormat.addInputPath(job2, input);
		String hstr = middledir + "/htmpempty/" + itr;
		FileOutputFormat.setOutputPath(job2, new Path(hstr));
		job2.waitForCompletion(true);
	}

	public static void computeW(Path input, String middledir, int k, int itr, int m, int n, int rp, int cp) throws IOException,
		InterruptedException, ClassNotFoundException {
		//job1
		Configuration conf = new Configuration();
		String hpath = middledir + "/hmatrix/" + itr;
		conf.set("middledir", middledir);
		//conf.setBoolean("mapred.output.compression.codec", true);
		//conf.setClass("mapred.output.compression.codec", GzipCodec.class,CompressionCodec.class);
		//conf.setInt("mapred.inmem.merge.threshold",0);
		//conf.setFloat("mapred.job.reduce.input.buffer.percent", 1);
		conf.set("mapred.child.java.opts", "-Xmx400m");
		//conf.setInt("io.sort.mb", 200);
		conf.setInt("io.sort.factor", 100);
		//conf.setInt("mapred.reduce.parallel.copies",10);
		conf.setInt("io.file.buffer.size", 128);
		conf.set("k", k + "");
		conf.set("itr", itr + "");
		conf.set("M", m + "");
		conf.set("N", n + "");
		conf.set("RP", rp + "");
		conf.set("CP", cp + "");
		Job job1 = new Job(conf, "computingHH"+itr);
		job1.setJarByClass(UpdateW.class);
		job1.setNumReduceTasks(1);
		job1.setJobName("computingHH"+itr);
		job1.setOutputKeyClass(IntWritable.class);
		job1.setOutputValueClass(Text.class);
		job1.setMapperClass(UpdateW.wmap1.class);
		job1.setReducerClass(UpdateW.wreduce1.class);
		job1.setInputFormatClass(NonSplittableTextInputFormat.class);
		job1.setOutputFormatClass(TextOutputFormat.class);


		FileInputFormat.addInputPath(job1, new Path(hpath));
		String outstr = middledir + "/hh/" + itr;
		FileOutputFormat.setOutputPath(job1, new Path(outstr));
		job1.waitForCompletion(true);


		//job2
		//avoid duplicated output files
		conf.setBoolean("mapred.map.tasks.speculative.execution", false);
		conf.setBoolean("mapred.reduce.tasks.speculative.execution", false);
		
		Job job2 = new Job(conf, "updateW"+itr);
		job2.setJarByClass(UpdateW.class);
		job2.setJobName("updateW"+itr);
		job2.setMapOutputKeyClass(IntPair.class);
		job2.setMapOutputValueClass(Text.class);
		job2.setOutputKeyClass(IntWritable.class);
		job2.setOutputValueClass(Text.class);
		job2.setNumReduceTasks(4);
		job2.setMapperClass(UpdateW.wmap2.class);
		job2.setReducerClass(UpdateW.wreduce2.class);
		//job2.setCombinerClass(hreduce2.class);
		job2.setInputFormatClass(NonSplittableTextInputFormat.class);
		job2.setOutputFormatClass(TextOutputFormat.class);


		//secondary sort
		job2.setPartitionerClass(FirstPartitioner.class);
		job2.setGroupingComparatorClass(GroupComparator.class);
		job2.setSortComparatorClass(KeyComparator.class);

		FileInputFormat.addInputPath(job2, input);
		String hstr = middledir + "/wtmpempty/" + itr;
		FileOutputFormat.setOutputPath(job2, new Path(hstr));
		job2.waitForCompletion(true);
	}

	
	public static void computeHfre(String input, String middledir, int k, int itr, int m, int n, int rp, int cp, int updateblocknum) throws IOException,
			InterruptedException, ClassNotFoundException {
		//job1
		Configuration conf = new Configuration();
		conf.set("middledir", middledir);
		//conf.setBoolean("mapred.output.compression.codec", true);
		//conf.setClass("mapred.output.compression.codec", GzipCodec.class,CompressionCodec.class);
		//conf.setInt("mapred.inmem.merge.threshold",0);
		//conf.setFloat("mapred.job.reduce.input.buffer.percent", 1);
		conf.set("mapred.child.java.opts", "-Xmx400m");
		//conf.setInt("io.sort.mb", 200);
		conf.setInt("io.sort.factor", 100);
		//conf.setInt("mapred.reduce.parallel.copies",10);
		conf.setInt("io.file.buffer.size", 128);
		conf.set("k", k + "");
		conf.set("itr", itr + "");
		conf.set("M", m + "");
		conf.set("N", n + "");
		conf.set("RP", rp + "");
		conf.set("CP", cp + "");
		conf.set("updatebnum", updateblocknum + "");
		Job job1 = new Job(conf, "computingWW"+itr);
		job1.setJarByClass(UpdateHfre.class);
		job1.setNumReduceTasks(1);
		job1.setJobName("computingWW"+itr);
		job1.setOutputKeyClass(IntWritable.class);
		job1.setOutputValueClass(Text.class);
		job1.setMapperClass(UpdateHfre.hmap1.class);
		job1.setReducerClass(UpdateHfre.hreduce1.class);
		job1.setInputFormatClass(NonSplittableTextInputFormat.class);
		job1.setOutputFormatClass(TextOutputFormat.class);
		
		FileSystem fs = FileSystem.get(conf);
		
		int start = 0;
		for(int i=0; i<updateblocknum; i++) // add updated w blocks into input
		{
			start = ((itr-3)*updateblocknum+i)%rp; // from iteration 3, start to use frequent updates (start index set to be 0)
			Path wpath = new Path(middledir + "/wmatrix/" + (itr-1) + "/w" + start);
			int begin = itr - 2;
			while(!fs.exists(wpath))
			{
				wpath = new Path(middledir + "/wmatrix/" + begin + "/w" + start);
				begin--;
			}
			FileInputFormat.addInputPath(job1, wpath);
		}
		
		String outstr = middledir + "/ww/" + itr;
		FileOutputFormat.setOutputPath(job1, new Path(outstr));
		job1.waitForCompletion(true);

		
		//job2
		//avoid duplicated output files
		conf.setBoolean("mapred.map.tasks.speculative.execution", false);
		conf.setBoolean("mapred.reduce.tasks.speculative.execution", false);
		
		Job job2 = new Job(conf, "updateHfre"+itr);
		job2.setJarByClass(UpdateHfre.class);
		job2.setJobName("updateHfre"+itr);
		job2.setMapOutputKeyClass(IntPair.class);
		job2.setMapOutputValueClass(Text.class);
		job2.setOutputKeyClass(IntWritable.class);
		job2.setOutputValueClass(Text.class);
		job2.setNumReduceTasks(4);
		job2.setMapperClass(UpdateHfre.hmap2.class);
		job2.setReducerClass(UpdateHfre.hreduce2.class);
		//job2.setCombinerClass(hreduce2.class);
		job2.setInputFormatClass(NonSplittableTextInputFormat.class);
		job2.setOutputFormatClass(TextOutputFormat.class);
		
		//secondary sort
		job2.setPartitionerClass(FirstPartitioner.class);
		job2.setGroupingComparatorClass(GroupComparator.class);
		job2.setSortComparatorClass(KeyComparator.class);
		
		start = 0;
		for(int i=0; i<updateblocknum; i++) // add updated w blocks into input
			for(int j=0; j<cp; j++)
			{
				start = ((itr-3)*updateblocknum+i)%rp; // from iteration 3, start to use frequent updates (start index set to be 0)
				FileInputFormat.addInputPath(job2, new Path(input + "/" + j + "-" + start + "-" + j));
			}
		String hstr = middledir + "/htmpempty/" + itr;
		FileOutputFormat.setOutputPath(job2, new Path(hstr));
		job2.waitForCompletion(true);
	}
	
	public static void computeWfre(String input, String middledir, int k, int itr, int m, int n, int rp, int cp, int updateblocknum) throws IOException,
	InterruptedException, ClassNotFoundException {
		//job1
		Configuration conf = new Configuration();
		String hpath = middledir + "/hmatrix/" + itr;
		conf.set("middledir", middledir);
		//conf.setBoolean("mapred.output.compression.codec", true);
		//conf.setClass("mapred.output.compression.codec", GzipCodec.class,CompressionCodec.class);
		//conf.setInt("mapred.inmem.merge.threshold",0);
		//conf.setFloat("mapred.job.reduce.input.buffer.percent", 1);
		conf.set("mapred.child.java.opts", "-Xmx400m");
		//conf.setInt("io.sort.mb", 200);
		conf.setInt("io.sort.factor", 100);
		//conf.setInt("mapred.reduce.parallel.copies",10);
		conf.setInt("io.file.buffer.size", 128);
		conf.set("k", k + "");
		conf.set("itr", itr + "");
		conf.set("M", m + "");
		conf.set("N", n + "");
		conf.set("RP", rp + "");
		conf.set("CP", cp + "");
		Job job1 = new Job(conf, "computingHH"+itr);
		job1.setJarByClass(UpdateWfre.class);
		job1.setNumReduceTasks(1);
		job1.setJobName("computingHH"+itr);
		job1.setOutputKeyClass(IntWritable.class);
		job1.setOutputValueClass(Text.class);
		job1.setMapperClass(UpdateWfre.wmap1.class);
		job1.setReducerClass(UpdateWfre.wreduce1.class);
		job1.setInputFormatClass(NonSplittableTextInputFormat.class);
		job1.setOutputFormatClass(TextOutputFormat.class);
		
		
		FileInputFormat.addInputPath(job1, new Path(hpath));
		String outstr = middledir + "/hh/" + itr;
		FileOutputFormat.setOutputPath(job1, new Path(outstr));
		job1.waitForCompletion(true);
	
	
		//job2
		//avoid duplicated output files
		conf.setBoolean("mapred.map.tasks.speculative.execution", false);
		conf.setBoolean("mapred.reduce.tasks.speculative.execution", false);
				
		Job job2 = new Job(conf, "updateWfre"+itr);
		job2.setJarByClass(UpdateWfre.class);
		job2.setJobName("updateWfre"+itr);
		job2.setMapOutputKeyClass(IntPair.class);
		job2.setMapOutputValueClass(Text.class);
		job2.setOutputKeyClass(IntWritable.class);
		job2.setOutputValueClass(Text.class);
		job2.setNumReduceTasks(4);
		job2.setMapperClass(UpdateWfre.wmap2.class);
		job2.setReducerClass(UpdateWfre.wreduce2.class);
		//job2.setCombinerClass(hreduce2.class);
		job2.setInputFormatClass(NonSplittableTextInputFormat.class);
		job2.setOutputFormatClass(TextOutputFormat.class);
		
		
		//secondary sort
		job2.setPartitionerClass(FirstPartitioner.class);
		job2.setGroupingComparatorClass(GroupComparator.class);
		job2.setSortComparatorClass(KeyComparator.class);
		
		int start = 0;
		for(int i=0; i<updateblocknum; i++) // add updated w blocks into input
			for(int j=0; j<cp; j++)
			{
				start = ((itr-2)*updateblocknum+i)%rp; // from iteration 2, start to use frequent updates (start index set to be 0)
				FileInputFormat.addInputPath(job2, new Path(input + "/" + j + "-" + start + "-" + j));
			}
		String hstr = middledir + "/wtmpempty/" + itr;
		FileOutputFormat.setOutputPath(job2, new Path(hstr));
		job2.waitForCompletion(true);
	}
	
	private static void checkPath(Path fileout1, FileSystem fs)
			throws IOException {
		if (fs.exists(fileout1))
			fs.delete(fileout1, true);
	}
	
	public static double runprogram(int m, int n, int iteration, int k, double range, String input, Path outpath, int rp, int cp, int updateblock) throws Exception{
		
		Path inputpath = new Path(input);
		FileSystem fs = inputpath.getFileSystem(new Configuration());
		Path middledir = new Path("hdfs:///tmpresult");
		checkPath(middledir, fs);
		fs.mkdirs(middledir);
		Path matrix=new Path(middledir,"matrix");
		if (!fs.exists(matrix)){
			fs.mkdirs(matrix);
		}
		
		String strdir = "mytempresult";
		
		Initialize_randommatrix.initW_hdfs(k, m, n, range, middledir, strdir, rp);
		Initialize_randommatrix.initH_hdfs(k, m, n, range, middledir, strdir, cp);
		
		long start = System.currentTimeMillis();
		long end = 0;
		long time = 0;
		double totaltime = 0.0;
		
		computeH(inputpath, strdir, k, 1, m, n, rp, cp);
		computeW(inputpath, strdir, k, 1, m, n, rp, cp);
		end = System.currentTimeMillis();
		time = end - start;
		totaltime = (double)time/1000;
		System.out.println("1 passed time:" + totaltime);
		computeH(inputpath, strdir, k, 2, m, n, rp, cp);
		computeWfre(input, strdir, k, 2, m, n, rp, cp, updateblock);
		end = System.currentTimeMillis();
		time = end - start;
		totaltime = (double)time/1000;
		System.out.println("2 passed time:" + totaltime);
		
		for (int i=3; i<=iteration; i++) {
			computeHfre(input, strdir, k, i, m, n, rp, cp, updateblock);
			computeWfre(input, strdir, k, i, m, n, rp, cp, updateblock);
			end = System.currentTimeMillis();
			time = end - start;
			totaltime = (double)time/1000;
			System.out.println(i + " passed time:" + totaltime);
		}
	
		end = System.currentTimeMillis();
		time = end - start;
		totaltime = (double)time/1000;
		return totaltime;
		
	}
	
	public static void main(String[] args) throws Exception {
		if (args.length != 10) {
            System.err.println("Usage: MainNMF <input> <output> <m> <n> <rp> <cp> <k> <itrlimit> <initial_range> <updateblock>");
            System.exit(2);
        }
		
		String input = args[0];
	    String output = args[1];
	    int m  = Integer.parseInt(args[2]);
	    int n  = Integer.parseInt(args[3]);
	    int rp  = Integer.parseInt(args[4]);
	    int cp  = Integer.parseInt(args[5]);
	    int k = Integer.parseInt(args[6]);
	    int itrlimit = Integer.parseInt(args[7]);
	    
	    double range = Double.parseDouble(args[8]);
	    int updateblock = Integer.parseInt(args[9]);
	       
		Path outpath=new Path(output);
		double time=runprogram(m, n, itrlimit, k, range, input, outpath, rp, cp, updateblock);
		System.out.print("time:" + time + "\n");
	}
	
}
