package frequentupdate;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileSplit;

import auxstructure.IntPair;

public class UpdateHfre {

	//compute W(T)*W
	public static class hmap1 extends
	Mapper<LongWritable, Text, IntWritable, Text> {
		
		private static double[][] ww;
		private int k = 10;
		
		@Override
		protected void setup(Context context) throws IOException,
				InterruptedException {
			
			k = Integer.parseInt(context.getConfiguration().get("k"));
			
			ww = new double[k][k];
			for(int i=0; i<k; i++)
				for(int j=0; j<k; j++)
				{
					ww[i][j] = 0;
				}
		}
		
		public void map(LongWritable key, Text value, Context context)
				throws IOException, InterruptedException {
			
			String[] vals = value.toString().split(",");
			double[] dvalues = new double[vals.length];
			for(int i=0; i<vals.length; i++)
			{
				dvalues[i] = Double.parseDouble(vals[i]);
			}
			
			for(int i=0; i<vals.length; i++)
				for(int j=0; j<vals.length; j++)
				{
					ww[i][j] += dvalues[i]*dvalues[j];
				}
		} 
		
		@Override
		protected void cleanup(Context context) throws IOException,
				InterruptedException {
			
			int itr = Integer.parseInt(context.getConfiguration().get("itr"));
			String mdir = context.getConfiguration().get("middledir");
			//String inputfileName = context.getConfiguration().get("map.input.file");			
			
			FileSplit fileSplit = (FileSplit)context.getInputSplit();
			String inputfileName = fileSplit.getPath().getName();
			
			String[] filestring = inputfileName.split("/");
			String blockid = filestring[filestring.length-1].substring(1);

            FSDataOutputStream out = null;
            BufferedWriter writer = null;
            try {
                    Path file = new Path(mdir + "/individualww/" + itr + "/" + blockid);
                    FileSystem fs = file.getFileSystem(new Configuration());
                    out = fs.create(file);
                    writer = new BufferedWriter(new OutputStreamWriter(out));
            } catch (IOException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
            }
            
			for(int i=0; i<k; i++)
			{
				String result = "";
				for(int j=0; j<k; j++)
				{
					if(j != k-1)
						result += ww[i][j] + ",";
					else
						result += ww[i][j];					
				}
				
				writer.write(i + "\t" +  result + "\n");
				
				context.write(new IntWritable(i), new Text(result));
			}
			
			try {
				writer.flush();
                writer.close();
                out.close();
			} catch (IOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
			}
		}
	}
	
	public static class hreduce1 extends
	Reducer<IntWritable, Text, IntWritable, Text> {
		
		private static double[][] ww;
		private int k = 10;
		private int itr = 0;
		
		@Override
		protected void setup(Context context) throws IOException,
				InterruptedException {
			k = Integer.parseInt(context.getConfiguration().get("k"));
			itr = Integer.parseInt(context.getConfiguration().get("itr"));
			
			ww = new double[k][k];
			for(int i=0; i<k; i++)
				for(int j=0; j<k; j++)
				{
					ww[i][j] = 0;
				}
			
			// load previous iteration result
			String mdir = context.getConfiguration().get("middledir");
			
			String wwpath = mdir + "/ww/" + (itr-1) + "/part-r-00000";
			
			Path file = new Path(wwpath);
			FileSystem fs = file.getFileSystem(new Configuration());
			FSDataInputStream ds = fs.open(file);
			BufferedReader reader = new BufferedReader(new InputStreamReader(ds));
			
			String line;
			while((line = reader.readLine()) != null)
			{
				String[] kvs  = line.split("\t");
				int index = Integer.parseInt(kvs[0]);
				String[] fields = kvs[1].split(",");
				
				for(int j=0; j<k; j++)
				{
					ww[index][j] += Double.parseDouble(fields[j]);
				}
			}
			reader.close();
			ds.close();
			
			int updateblocknum = Integer.parseInt(context.getConfiguration().get("updatebnum"));
			int rp = Integer.parseInt(context.getConfiguration().get("RP"));
			//int stepsperiteration = rp/updateblocknum;
			
			int start = 0;
			for(int i=0; i<updateblocknum; i++) // deduct the updated blocks
			{
				start = ((itr-3)*updateblocknum+i)%rp; // updated w blocks
				wwpath = mdir + "/individualww/" + (itr-1) + "/" + start;
				file = new Path(wwpath);
				//look for the right path
				int begin = itr - 2;
				while(!fs.exists(file))
				{
					wwpath = mdir + "/individualww/" + begin + "/" + start;
					file = new Path(wwpath);
					begin--;
				}
				ds = fs.open(file);
				reader = new BufferedReader(new InputStreamReader(ds));
				
				while((line = reader.readLine()) != null)
				{
					String[] kvs  = line.split("\t");
					int index = Integer.parseInt(kvs[0]);
					String[] fields = kvs[1].split(",");
					
					for(int j=0; j<k; j++)
					{
						ww[index][j] -= Double.parseDouble(fields[j]);
					}
				}
				reader.close();
				ds.close();
			}
			
		}

		public void reduce(IntWritable key, Iterable<Text> values,
				Context context) throws IOException, InterruptedException {
			
			int index = key.get();
	
			for (Text value : values) {
				String[] fields = value.toString().split(",");
				
				for(int j=0; j<k; j++)
				{
					ww[index][j] += Double.parseDouble(fields[j]);
				}
			}
			String result = "";
			for (int counter = 0; counter < k; counter++) {
				result += ww[index][counter];
				if (counter != k-1)
					result += ",";
			}

			context.write(new IntWritable(index), new Text(result));
		}
	}
	
	//compute W(T)*A
	public static class hmap2 extends
	Mapper<LongWritable, Text, IntPair, Text> {
			
		private static double[][] w;
		private static double[][] y;
		
		private int k;
		private int M;
		private int N;
		private int RP; // rows are split in to RP parts
		private int CP; // columns are split in to CP parts
		private int itr; // iteration
		
		private int RperPart;
		private int CperPart;
		
		private boolean loadw = false;
		private int highrow = 0;
		private int highcolumn = 0;
		
		private int startrow = 0;
		private int startcolumn = 0;
		
		private boolean initialized = false;
		
		private int recordlimit = 0;
		private int recordlimit2 = 0;
		
		/*
		public int getRowPartNum(int i)
		{
			int num = 0;
			
			num = i/(M/RP);
			
			return num;
		}
		
		public int getColumnPartNum(int j)
		{
			int num = 0;
			
			num = j/(N/CP);
			
			return num;
		}
		*/
		
		@Override
		protected void setup(Context context) throws IOException,
				InterruptedException {
			
			k = Integer.parseInt(context.getConfiguration().get("k"));
			M = Integer.parseInt(context.getConfiguration().get("M"));
			N = Integer.parseInt(context.getConfiguration().get("N"));
			RP = Integer.parseInt(context.getConfiguration().get("RP"));
			CP = Integer.parseInt(context.getConfiguration().get("CP"));
			itr = Integer.parseInt(context.getConfiguration().get("itr"));
				
			if(M%RP == 0)
				RperPart = M/RP;
			else
				RperPart = M/RP + 1;
			
			if(N%CP == 0)
				CperPart = N/CP;
			else
				CperPart = N/CP + 1;
			
			w = new double[RperPart][k];
			y = new double[k][CperPart];
			//need to resize h
			
			for(int i=0; i<RperPart; i++)
				for(int j=0; j<k; j++)
				{
					w[i][j] = 0;
				}
			
			for(int i=0; i<k; i++)
				for(int j=0; j<CperPart; j++)
				{
					y[i][j] = 0;
				}
			
		}
			
		public void map(LongWritable key, Text value, Context context)
				throws IOException, NumberFormatException, InterruptedException {

//			String delim = new String();
//			if (value.toString().contains("\t")) {
//				delim = "\t";
//			} else if (value.toString().contains("::")) {
//				delim = "::";
//			}
//			String[] vals = value.toString().split(delim);
			
			int row = 0;
			int column = 0;
			double element = 0;
			
			String[] fields = value.toString().split(",");
			
			//if(fields.length != 3)
			//	System.out.println(":" + value.toString() + ":");
			
			row = Integer.parseInt(fields[0]);
			column = Integer.parseInt(fields[1]);
			element = Double.parseDouble(fields[2]);
			
			
			if(!initialized)
			{
				initialized = true;
				
				highrow = row/RperPart;
				highcolumn = column/CperPart;
				
				//System.out.println("column: " + column + ";   CperPart: " + CperPart);
				//System.out.println("in map, highcolumn:" + highcolumn);
				
				startrow = highrow * RperPart;
				startcolumn = highcolumn * CperPart;
			}
						
			if(!loadw)
			{
				loadw = true;
				
				try {
					String mdir = context.getConfiguration().get("middledir");
					
					String wpath = mdir + "/wmatrix/" + (itr-1) + "/w" + highrow;
					//generate final path using highrow
					Path file = new Path(wpath);
					FileSystem fs = file.getFileSystem(new Configuration());
					FSDataInputStream ds = fs.open(file);
					BufferedReader reader = new BufferedReader(new InputStreamReader(ds));
					
					String line;
					int count = 0;
					while((line = reader.readLine()) != null)
					{
						String tmpfields[] = line.split(",");
						
						double ev = 0;
						for(int i=0; i<tmpfields.length; i++)
						{
							ev = Double.parseDouble(tmpfields[i]);
							w[count][i] = ev;
						}
						count++;
					}
					reader.close();
					ds.close();
				} catch (Exception e) {

				}
			}
			
			if(recordlimit < 100 && Double.compare(element, Double.NaN) == 0 )
			{
				System.out.println("Double.NaN by: " + value.toString());
				recordlimit++;
			}
			
			for(int j=0; j<k; j++)
			{
				y[j][column-startcolumn] += element * w[row-startrow][j];  
			}
			
//			for(int i=0; i<vals.length; i++)
//			{
//				fields = vals[i].split(",");
//				row = Integer.parseInt(fields[0]);
//				column = Integer.parseInt(fields[1]);
//				element = Double.parseDouble(fields[2]);
//				//row and column can be used to determine to load which H
//				
//				for(int j=0; j<k; j++)
//				{
//					y[j][column-startcolumn] += element * w[row-startrow][j];  
//				}
//			}
		}
			
			@Override
			protected void cleanup(Context context) throws IOException,
					InterruptedException {
				
				String mdir = context.getConfiguration().get("middledir");

	            FSDataOutputStream out = null;
	            BufferedWriter writer = null;
	            try {
	                    Path file = new Path(mdir + "/individualwta/" + itr + "/" + highrow + "-" + highcolumn);
	                    FileSystem fs = file.getFileSystem(new Configuration());
	                    out = fs.create(file);
	                    writer = new BufferedWriter(new OutputStreamWriter(out));
	            } catch (IOException e) {
	                    // TODO Auto-generated catch block
	                    e.printStackTrace();
	            }
	            
				//System.out.println("in cleanup, highcolumn:" + highcolumn);
				for(int j=0; j<CperPart; j++)
				{	
					String result = "";
					for(int i=0; i<k; i++)
					{
						if(recordlimit2 < 100 && Double.compare(y[i][j], Double.NaN) == 0 )
						{
							System.out.println("y[i][j]: " + y[i][j]);
							recordlimit2++;
						}
						
						if(i != k-1)
							result += y[i][j] + ",";
						else
							result += y[i][j];
					}
					
					writer.write(j + "\t" +  result + "\n");
					
					context.write(new IntPair(highcolumn, j), new Text(result));
				}
				
				try {
					writer.flush();
	                writer.close();
	                out.close();
				} catch (IOException e) {
	                // TODO Auto-generated catch block
	                e.printStackTrace();
				}
			}
		}
		
		public static class hreduce2 extends
		Reducer<IntPair, Text, IntWritable, Text> {
			
			private int k = 10;
			private int M;
			private int N = 100;
			private int RP; // rows are split in to RP parts
			private int CP; // columns are split in to CP parts
			private int itr; // iteration
			
			private int RperPart;
			private int CperPart;
			
			private static double[][] ww;
			
			private BufferedReader readerinreduce = null;
			private FSDataInputStream hds = null;
			private boolean initilized = false;
			
			private FSDataOutputStream out = null;
			private BufferedWriter writer = null;
			
			private int recordlimit = 0;
			private int recordlimit2 = 0;
			
			@Override
			protected void setup(Context context) throws IOException,
					InterruptedException {
				k = Integer.parseInt(context.getConfiguration().get("k"));
				M = Integer.parseInt(context.getConfiguration().get("M"));
				N = Integer.parseInt(context.getConfiguration().get("N"));
				itr = Integer.parseInt(context.getConfiguration().get("itr"));
				RP = Integer.parseInt(context.getConfiguration().get("RP"));
				CP = Integer.parseInt(context.getConfiguration().get("CP"));
				
				if(M%RP == 0)
					RperPart = M/RP;
				else
					RperPart = M/RP + 1;
				
				if(N%CP == 0)
					CperPart = N/CP;
				else
					CperPart = N/CP + 1;
				
				ww = new double[k][k];
				
				try {
					String mdir = context.getConfiguration().get("middledir");
					String wwpath = mdir + "/ww/" + itr + "/part-r-00000";
					Path file = new Path(wwpath);
					FileSystem fs = file.getFileSystem(new Configuration());
					FSDataInputStream ds = fs.open(file);
					BufferedReader reader = new BufferedReader(new InputStreamReader(ds));
					
					String line;
					while((line = reader.readLine()) != null)
					{
						int index = line.indexOf("\t");
						if(index != -1)
						{
							String cidstr = line.substring(0, index);
							int cid = Integer.parseInt(cidstr);
							
							String values = line.substring(index+1);
							String fields[] = values.split(",");
							
							double ev = 0;
							for(int i=0; i<fields.length; i++)
							{
								ev = Double.parseDouble(fields[i]);
								ww[cid][i] = ev;
							}
						}
					}
					reader.close();
					ds.close();
					
				} catch (Exception e) {

				}
				
			}

			public void reduce(IntPair key, Iterable<Text> values,
					Context context) throws IOException, InterruptedException {
				
				int bigj = key.getFirst();
				int smallj = key.getSecond();
				
				//System.out.println("bigj:" + bigj);
				
				//need to use bigj to decide which file
				String mdir = context.getConfiguration().get("middledir");
				
				String hpath = mdir + "/hmatrix/" + (itr-1) + "/h" + bigj;
				
				Path hfile = new Path(hpath);
				FileSystem hdfs = hfile.getFileSystem(new Configuration());
				hds = hdfs.open(hfile);
				readerinreduce = new BufferedReader(new InputStreamReader(hds));
				
				FileSystem fs = FileSystem.get(context.getConfiguration());
				Path outPath = new Path(mdir + "/hmatrix/" + itr + "/h" + bigj);
				out = fs.create(outPath);
				writer = new BufferedWriter(new OutputStreamWriter(out));
				
				Path outwtaPath = new Path(mdir + "/totalwta/" + itr + "/" + bigj);
				FSDataOutputStream wtaout = fs.create(outwtaPath);
				BufferedWriter wtawriter = new BufferedWriter(new OutputStreamWriter(wtaout));
				
				double[][] Z = new double[k][CperPart];
				double[] Hj = new double[k];
				int column = 0;
				
				
				//initialize Z
				Path wtapath = new Path(mdir + "/totalwta/" + (itr-1) + "/" + bigj);
				FSDataInputStream ds = fs.open(wtapath);
				BufferedReader reader = new BufferedReader(new InputStreamReader(ds));
				
				String line;
				while((line = reader.readLine()) != null)
				{
					String[] kvs  = line.split("\t");
					int index = Integer.parseInt(kvs[0]);
					String[] fields = kvs[1].split(",");
					
					for(int j=0; j<k; j++)
					{
						Z[j][index] += Double.parseDouble(fields[j]);
					}
				}
				
				reader.close();
				ds.close();
				
				System.out.println("load total wta, z[0][0]:" + Z[0][0]);
				
				int updateblocknum = Integer.parseInt(context.getConfiguration().get("updatebnum"));
				
				int start = 0;
				for(int i=0; i<updateblocknum; i++) // deduct the updated blocks
				{
					start = ((itr-3)*updateblocknum+i)%RP; // updated w blocks
					wtapath = new Path(mdir + "/individualwta/" + (itr-1) + "/" + start + "-" + bigj);
					int begin = itr - 2;
					while(!fs.exists(wtapath))
					{
						wtapath = new Path(mdir + "/individualwta/" + begin + "/" + start + "-" + bigj);
						begin--;
					}
					ds = fs.open(wtapath);
					reader = new BufferedReader(new InputStreamReader(ds));
					
					while((line = reader.readLine()) != null)
					{
						String[] kvs  = line.split("\t");
						int index = Integer.parseInt(kvs[0]);
						String[] fields = kvs[1].split(",");
						
						for(int j=0; j<k; j++)
						{
							Z[j][index] -= Double.parseDouble(fields[j]);
						}
					}
				}
				reader.close();
				ds.close();
				
				System.out.println("deduct old values, z[0][0]:" + Z[0][0]);
		
				//all values for the same J
				for (Text value : values) {
					column = key.getSecond();
					if(column == smallj)
					{
						String[] fields = value.toString().split(",");
						for(int i=0; i<k; i++)
						{
							Z[i][column] += Double.parseDouble(fields[i]);
						}
					}
					else
					{
						//get all results of Z[]
						String wtastr = "";
						for(int r=0; r<k; r++)
						{
							if(r != k-1)
								wtastr += Z[r][column-1] + ",";
							else
								wtastr += Z[r][column-1];
						}
						wtawriter.write((column-1) + "\t" + wtastr + "\n");
						
						if((line = readerinreduce.readLine()) != null)
						{
							String fields[] = line.split(",");
							
							double ev = 0;
							for(int i=0; i<fields.length; i++)
							{
								ev = Double.parseDouble(fields[i]);
								Hj[i] = ev;
							}
						}
						double[] tempresult = new double[k];
						
						double tempvalue = 0;
						for(int r=0; r<k; r++)
						{
							tempvalue = 0;
							for(int j=0; j<k; j++)
							{
								tempvalue += ww[r][j] * Hj[j] ;
							}
							tempresult[r] = tempvalue;
						}
						double[] result = new double[k];
						String outputstr = "";
						for(int cl=0; cl<k; cl++)
						{
							if(Double.compare(Hj[cl], 0.0) == 0 || Double.compare(Z[cl][column-1], 0.0) == 0 || Double.compare(tempresult[cl], 0.0) == 0)
								result[cl] = 0;
							else
								result[cl] = Hj[cl]*(Z[cl][column-1]/tempresult[cl]);
							
							if(recordlimit < 100 && (Double.compare(result[cl], Double.NaN) == 0 || Double.compare(result[cl], 0.0) == 0))
							{
								System.out.println(cl + " : Hj[cl]: " + Hj[cl] + "  Z[cl]: " + Z[cl] + "  tempresult[cl]: " + tempresult[cl]);
								recordlimit++;
							}
							
							if(cl != k-1)
								outputstr += result[cl] + ",";
							else
								outputstr += result[cl];
						}	
						writer.write(outputstr + "\n"); // if one reducer handle multiple bigj, need to write files directly						
						
						
						smallj = column;
						String[] fields = value.toString().split(",");
						for(int i=0; i<k; i++)
						{
							Z[i][column] += Double.parseDouble(fields[i]);
						}
					}
				}
				
				//last column
				if((line = readerinreduce.readLine()) != null)
				{
					String fields[] = line.split(",");
					
					double ev = 0;
					for(int i=0; i<fields.length; i++)
					{
						ev = Double.parseDouble(fields[i]);
						Hj[i] = ev;
					}
				}
				
				//get all results of Z[]
				String wtastr = "";
				for(int r=0; r<k; r++)
				{
					if(r != k-1)
						wtastr += Z[r][column] + ",";
					else
						wtastr += Z[r][column];
				}
				wtawriter.write(column + "\t" + wtastr + "\n");
				
				double[] tempresult = new double[k];
				
				double tempvalue = 0;
				for(int r=0; r<k; r++)
				{
					tempvalue = 0;
					for(int j=0; j<k; j++)
					{
						tempvalue += ww[r][j] * Hj[j] ;
					}
					tempresult[r] = tempvalue;
				}
				double[] result = new double[k];
				String outputstr = "";
				for(int cl=0; cl<k; cl++)
				{
					if(Double.compare(Hj[cl], 0.0) == 0 || Double.compare(Z[cl][column], 0.0) == 0 || Double.compare(tempresult[cl], 0.0) == 0)
						result[cl] = 0;
					else
						result[cl] = Hj[cl]*(Z[cl][column]/tempresult[cl]);
					
					if(cl != k-1)
						outputstr += result[cl] + ",";
					else
						outputstr += result[cl];
				}	
				writer.write(outputstr + "\n");
				
				readerinreduce.close();
				hds.close();
				
				wtawriter.flush();
				wtawriter.close();
				wtaout.close();
				
				writer.flush();
				writer.close();
				out.close();
			}
		}
		
		//compute H
		/*
		public static class hmap3 extends
		Mapper<LongWritable, Text, LongWritable, Text> {
				
			private static double[][] ww; 
			private int k=10;
			private BufferedReader readerinmap = null;
			private FSDataInputStream yds = null;
				
			@Override
			protected void setup(Context context) throws IOException,
					InterruptedException {
				
				k = Integer.parseInt(context.getConfiguration().get("k"));
				int N = Integer.parseInt(context.getConfiguration().get("N"));
					
				ww = new double[k][k];
				//need to resize h
				
				for(int i=0; i<k; i++)
					for(int j=0; j<k; j++)
					{
						ww[i][j] = 0;
					}
				
				try {
					
					String wwpath = context.getConfiguration().get("wwpath");
					Path file = new Path(wwpath);
					FileSystem fs = file.getFileSystem(new Configuration());
					FSDataInputStream ds = fs.open(file);
					BufferedReader reader = new BufferedReader(new InputStreamReader(ds));
					
					String line;
					while((line = reader.readLine()) != null)
					{
						int index = line.indexOf("\t");
						if(index != -1)
						{
							String cidstr = line.substring(0, index);
							int cid = Integer.parseInt(cidstr);
							
							String values = line.substring(index+1);
							String fields[] = values.split(",");
							
							double ev = 0;
							for(int i=0; i<fields.length; i++)
							{
								ev = Double.parseDouble(fields[i]);
								ww[i][cid] = ev;
							}
						}
					}
					reader.close();
					ds.close();
					
					String ypath = context.getConfiguration().get("ypath");
					Path yfile = new Path(ypath);
					FileSystem hdfs = yfile.getFileSystem(new Configuration());
					yds = hdfs.open(yfile);
					readerinmap = new BufferedReader(new InputStreamReader(yds));
					
				} catch (Exception e) {

				}
			}
				
			public void map(LongWritable key, Text value, Context context)
					throws IOException, NumberFormatException, InterruptedException {

				String delim = new String();
				if (value.toString().contains("\t")) {
					delim = "\t";
				} else if (value.toString().contains("::")) {
					delim = "::";
				}
				String[] vals = value.toString().split(delim);
				
				int row = (int)key.get();
				
				double[] Hj = new double[k];
				for(int i=0; i<vals.length; i++)
				{
					Hj[i] = Double.parseDouble(vals[i]);
				}
				
				double[] tempresult = new double[k];
				
				double tempvalue = 0;
				for(int r=0; r<k; r++)
				{
					tempvalue = 0;
					for(int j=0; j<vals.length; j++)
					{
						tempvalue += ww[r][j] * Hj[j] ;
					}
					tempresult[r] = tempvalue;
				}
				
				double[] Y = new double[k];
				String line;
				if((line = readerinmap.readLine()) != null)
				{
					int index = line.indexOf("\t");
					if(index != -1)
					{
						String cidstr = line.substring(0, index);
						int cid = Integer.parseInt(cidstr);
						
						String values = line.substring(index+1);
						String fields[] = values.split(",");
						
						double ev = 0;
						for(int i=0; i<fields.length; i++)
						{
							ev = Double.parseDouble(fields[i]);
							Y[i] = ev;
						}
					}
				}
				
				double[] result = new double[k];
				String outputstr = "";
				for(int column=0; column<k; column++)
				{
					result[column] = Hj[column]*(Y[column]/tempresult[column]);
					
					if(column != k-1)
						outputstr += result[column] + ",";
					else
						outputstr += result[column];
				}
				
				context.write(new LongWritable(row), new Text(outputstr));

			}
				
			@Override
			protected void cleanup(Context context) throws IOException,
						InterruptedException {
				readerinmap.close();
				yds.close();
			}

		}
			
		//no need reducer in this job
		public static class hreduce3 extends
		Reducer<LongWritable, Text, LongWritable, Text> {
				
			private static double[][] ah; 
			int k = 10;
			int M = 100;
				
			@Override
			protected void setup(Context context) throws IOException,
					InterruptedException {
				k = Integer.parseInt(context.getConfiguration().get("k"));
				M = Integer.parseInt(context.getConfiguration().get("M"));
					
			}

			public void reduce(LongWritable key, Iterable<Text> values,
					Context context) throws IOException, InterruptedException {
					
			}
		}
		*/
		
	
		public static void main(String[] args) {
		// TODO Auto-generated method stub

		}

}
