package frequentupdate;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;

import auxstructure.IntPair;


public class UpdateW {

	//compute H*H(T)
	public static class wmap1 extends
	Mapper<LongWritable, Text, IntWritable, Text> {
		
		private double[][] hh;
		private int k;
		private int testlimit = 0;
		private int testlimit2 = 0;
		private int testlimit3 = 0;
		private double oldvalue = 0;
		
		private int recordcount = 0;
		
		@Override
		protected void setup(Context context) throws IOException,
				InterruptedException {
			
			k = Integer.parseInt(context.getConfiguration().get("k"));
			
			hh = new double[k][k];
			for(int i=0; i<k; i++)
				for(int j=0; j<k; j++)
				{
					hh[i][j] = 0;
				}
		}
		
		public void map(LongWritable key, Text value, Context context)
				throws IOException, InterruptedException {
			
			recordcount++;
			
			if(testlimit < 10)
			{
				System.out.println(value.toString());
				testlimit++;
			}
			
			String[] vals = value.toString().split(",");
			double[] dvalues = new double[vals.length];
			for(int i=0; i<vals.length; i++)
			{
				dvalues[i] = Double.parseDouble(vals[i]);
				
				if(Double.compare(dvalues[i], Double.NaN) == 0 && testlimit3 < 100)
				{
					testlimit3++;
					System.out.println("Double.NaN by: " + value.toString());
				}
			}
			for(int i=0; i<vals.length; i++)
				for(int j=0; j<vals.length; j++)
				{
					hh[i][j] += dvalues[i]*dvalues[j];
				}
			
			if(testlimit2 < 200 &&  Double.compare(hh[0][0], Double.NaN) == 0 && hh[0][0] != oldvalue)
			{
				System.out.println("hh[0][0]: " + hh[0][0]);
				testlimit2++;
				oldvalue = hh[0][0];
			}
		}
		
		@Override
		protected void cleanup(Context context) throws IOException,
				InterruptedException {
			
			for(int i=0; i<k; i++)
			{
				String result = "";
				for(int j=0; j<k; j++)
				{
					System.out.println(hh[i][j]);
					
					if(j != k-1)
						result += hh[i][j] + ",";
					else
						result += hh[i][j];
				}
				
				context.write(new IntWritable(i), new Text(result));
			}
		}

	}
	
	public static class wreduce1 extends
	Reducer<IntWritable, Text, LongWritable, Text> {
		
		private double[][] hh; 
		private int k;
		
		@Override
		protected void setup(Context context) throws IOException,
				InterruptedException {
			
			k = Integer.parseInt(context.getConfiguration().get("k"));
			
			hh = new double[k][k];
			for(int i=0; i<k; i++)
				for(int j=0; j<k; j++)
				{
					hh[i][j] = 0;
				}
		}

		public void reduce(IntWritable key, Iterable<Text> values,
				Context context) throws IOException, InterruptedException {
			
			int index = key.get();
	
			for (Text value : values) {
				
				String[] fields = value.toString().split(",");
				for(int j=0; j<k; j++)
				{
					hh[index][j] += Double.parseDouble(fields[j]);
				}
			}
			
			String result = "";
			for (int counter = 0; counter < k; counter++) {
				System.out.println(hh[index][counter]);
				
				result += hh[index][counter];
				if (counter != k - 1)
					result += ",";
			}

			context.write(new LongWritable(index), new Text(result));
		}
	}
	
	//compute A*H(T)
	public static class wmap2 extends
	Mapper<LongWritable, Text, IntPair, Text> {
			
		private static double[][] h;
		private static double[][] y;
		
		private int k;
		private int M;
		private int N;
		
		private int RP; // rows are split in to RP parts
		private int CP; // columns are split in to CP parts
		private int itr; // iteration

		private int RperPart;
		private int CperPart;
		
		private boolean loadh = false;
		private int highrow = 0;
		private int highcolumn = 0;
		
		private int startrow = 0;
		private int startcolumn = 0;
		
		private boolean initialized = false;
			
		@Override
		protected void setup(Context context) throws IOException,
				InterruptedException {
			
			k = Integer.parseInt(context.getConfiguration().get("k"));
			M = Integer.parseInt(context.getConfiguration().get("M"));
			N = Integer.parseInt(context.getConfiguration().get("N"));
			RP = Integer.parseInt(context.getConfiguration().get("RP"));
			CP = Integer.parseInt(context.getConfiguration().get("CP"));
			itr = Integer.parseInt(context.getConfiguration().get("itr"));
			
			if(M%RP == 0)
				RperPart = M/RP;
			else
				RperPart = M/RP + 1;
			
			if(N%CP == 0)
				CperPart = N/CP;
			else
				CperPart = N/CP + 1;
				
			h = new double[k][CperPart];
			y = new double[RperPart][k];
			//need to resize h
			
			for(int i=0; i<k; i++)
				for(int j=0; j<CperPart; j++)
				{
					h[i][j] = 0;
				}
			
			for(int i=0; i<RperPart; i++)
				for(int j=0; j<k; j++)
				{
					y[i][j] = 0;
				}
		}
			
		public void map(LongWritable key, Text value, Context context)
				throws IOException, NumberFormatException, InterruptedException {
				
//			String delim = new String();
//			if (value.toString().contains("\t")) {
//				delim = "\t";
//			} else if (value.toString().contains("::")) {
//				delim = "::";
//			}
//			String[] vals = value.toString().split(delim);
			
			int row = 0;
			int column = 0;
			double element = 0;
			
			String[] fields = value.toString().split(",");
			row = Integer.parseInt(fields[0]);
			column = Integer.parseInt(fields[1]);
			element = Double.parseDouble(fields[2]);
			
			if(!initialized)
			{
				initialized = true;
				
				highrow = row/RperPart;
				highcolumn = column/CperPart;
				
				startrow = highrow * RperPart;
				startcolumn = highcolumn * CperPart;
			}
			
			if(!loadh)
			{
				loadh = true;
				
				try {
					String mdir = context.getConfiguration().get("middledir");
					
					String hpath = mdir + "/hmatrix/" + itr + "/h" + highcolumn;
					//generate final path using highcolumn
					Path file = new Path(hpath);
					FileSystem fs = file.getFileSystem(new Configuration());
					FSDataInputStream ds = fs.open(file);
					BufferedReader reader = new BufferedReader(new InputStreamReader(ds));
					
					String line;
					int count = 0;
					while((line = reader.readLine()) != null)
					{
						fields = line.split(",");
						
						double ev = 0;
						for(int i=0; i<fields.length; i++)
						{
							ev = Double.parseDouble(fields[i]);
							h[i][count] = ev;
						}
						count++;
					}
					reader.close();
					ds.close();
				} catch (Exception e) {

				}
			}
			
			for(int j=0; j<k; j++)
			{
				y[row-startrow][j] += element * h[j][column-startcolumn];  
			}
			
//			for(int i=0; i<vals.length; i++)
//			{
//				fields = vals[i].split(",");
//				row = Integer.parseInt(fields[0]);
//				column = Integer.parseInt(fields[1]);
//				element = Double.parseDouble(fields[2]);
//				//row and column can be used to determine to load which H
//				
//				for(int j=0; j<k; j++)
//				{
//					y[row-startrow][j] += element * h[j][column-startcolumn];  
//				}
//			}

		}
			
			@Override
			protected void cleanup(Context context) throws IOException,
					InterruptedException {
				
				for(int i=0; i<RperPart; i++)
				{	
					String result = "";
					for(int j=0; j<k; j++)
					{
						if(j != k-1)
							result += y[i][j] + ",";
						else
							result += y[i][j];
					}
					
					context.write(new IntPair(highrow, i), new Text(result));
				}
			}

	}
		
		public static class wreduce2 extends
		Reducer<IntPair, Text, IntWritable, Text> {
			
			int k = 10;
			int M = 100;
			private int itr; // iteration
			
			private static double[][] hh;
			
			private BufferedReader readerinreduce = null;
			private FSDataInputStream wds = null;
			private boolean initilized = false;
			
			private FSDataOutputStream out = null;
			private BufferedWriter writer = null;
			
			@Override
			protected void setup(Context context) throws IOException,
					InterruptedException {
				k = Integer.parseInt(context.getConfiguration().get("k"));
				//M = Integer.parseInt(context.getConfiguration().get("M"));
			    itr = Integer.parseInt(context.getConfiguration().get("itr"));
				
				hh = new double[k][k];
				
				try {
					String mdir = context.getConfiguration().get("middledir");
					String hhpath = mdir + "/hh/" + itr + "/part-r-00000";
					Path file = new Path(hhpath);
					FileSystem fs = file.getFileSystem(new Configuration());
					FSDataInputStream ds = fs.open(file);
					BufferedReader reader = new BufferedReader(new InputStreamReader(ds));
					
					String line;
					while((line = reader.readLine()) != null)
					{
						int index = line.indexOf("\t");
						if(index != -1)
						{
							String cidstr = line.substring(0, index);
							int cid = Integer.parseInt(cidstr);
							
							String values = line.substring(index+1);
							String fields[] = values.split(",");
							
							double ev = 0;
							for(int i=0; i<fields.length; i++)
							{
								ev = Double.parseDouble(fields[i]);
								hh[cid][i] = ev;
							}
						}
					}
					reader.close();
					ds.close();
					
				} catch (Exception e) {

				}
			}

			public void reduce(IntPair key, Iterable<Text> values,
					Context context) throws IOException, InterruptedException {
				
				int bigi = key.getFirst();
				int smalli = key.getSecond();
				
				
				//need to use bigi to decide which file
				String mdir = context.getConfiguration().get("middledir");
				
				String wpath = mdir + "/wmatrix/" + (itr-1) + "/w" + bigi;
				Path wfile = new Path(wpath);
				FileSystem hdfs = wfile.getFileSystem(new Configuration());
				wds = hdfs.open(wfile);
				readerinreduce = new BufferedReader(new InputStreamReader(wds));
				
				FileSystem fs = FileSystem.get(context.getConfiguration());
				Path outPath = new Path(mdir + "/wmatrix/" + itr + "/w" + bigi);
				out = fs.create(outPath);
				writer = new BufferedWriter(new OutputStreamWriter(out));
				
				//if(!initilized)
				//{
				//	initilized = true;
				//}
				
				double[] Z = new double[k];
				double[] Wi = new double[k];
				int row = 0;
				double dv = 0;
				String line;
				
				for(int i=0; i<k; i++)
				{
					Z[i] = 0;
				}
		
				for (Text value : values) {
					
					row = key.getSecond();
					if(row == smalli)
					{
						String[] fields = value.toString().split(",");
						for(int j=0; j<k; j++)
						{
							Z[j] += Double.parseDouble(fields[j]);
						}
					}
					else
					{
						if((line = readerinreduce.readLine()) != null)
						{
							String fields[] = line.split(",");
							
							double ev = 0;
							for(int i=0; i<fields.length; i++)
							{
								ev = Double.parseDouble(fields[i]);
								Wi[i] = ev;
							}
						}
						double[] tempresult = new double[k];
						
						double tempvalue = 0;
						for(int column=0; column<k; column++)
						{
							tempvalue = 0;
							for(int i=0; i<k; i++)
							{
								tempvalue += Wi[i] * hh[i][column];
							}
							tempresult[column] = tempvalue;
						}
						
						double[] result = new double[k];
						String outputstr = "";
						for(int column=0; column<k; column++)
						{
							if(Double.compare(Wi[column], 0.0) == 0 || Double.compare(Z[column], 0.0) == 0 || Double.compare(tempresult[column], 0.0) == 0)
								result[column] = 0;
							else
								result[column] = Wi[column]*(Z[column]/tempresult[column]);
							
							if(column != k-1)
								outputstr += result[column] + ",";
							else
								outputstr += result[column];
						}
						writer.write(outputstr + "\n"); // if one reducer handle multiple bigi, need to write files directly		
						
						smalli = row;
						String[] fields = value.toString().split(",");
						for(int j=0; j<k; j++)
						{
							Z[j] = Double.parseDouble(fields[j]);
						}
					}
				}
				
				//last row
				if((line = readerinreduce.readLine()) != null)
				{
					String fields[] = line.split(",");
					
					double ev = 0;
					for(int i=0; i<fields.length; i++)
					{
						ev = Double.parseDouble(fields[i]);
						Wi[i] = ev;
					}
				}
				double[] tempresult = new double[k];
				
				double tempvalue = 0;
				for(int column=0; column<k; column++)
				{
					tempvalue = 0;
					for(int i=0; i<k; i++)
					{
						tempvalue += Wi[i] * hh[i][column];
					}
					tempresult[column] = tempvalue;
				}
				
				double[] result = new double[k];
				String outputstr = "";
				for(int column=0; column<k; column++)
				{
					if(Double.compare(Wi[column], 0.0) == 0 || Double.compare(Z[column], 0.0) == 0 || Double.compare(tempresult[column], 0.0) == 0)
						result[column] = 0;
					else
						result[column] = Wi[column]*(Z[column]/tempresult[column]);
					
					if(column != k-1)
						outputstr += result[column] + ",";
					else
						outputstr += result[column];
				}	
				writer.write(outputstr + "\n");
				
				readerinreduce.close();
				wds.close();
				
				writer.flush();
				writer.close();
				out.close();
			}
		}
		
		//compute W
		/*
		public static class wmap3 extends
		Mapper<LongWritable, Text, LongWritable, Text> {
				
			private static double[][] hh; 
			private int k=10;
			private BufferedReader readerinmap = null;
			private FSDataInputStream yds = null;
				
			@Override
			protected void setup(Context context) throws IOException,
					InterruptedException {
				
				k = Integer.parseInt(context.getConfiguration().get("k"));
				int N = Integer.parseInt(context.getConfiguration().get("N"));
					
				hh = new double[k][k];
				//need to resize h
				
				for(int i=0; i<k; i++)
					for(int j=0; j<k; j++)
					{
						hh[i][j] = 0;
					}
				
				try {
					
					String hhpath = context.getConfiguration().get("hhpath");
					Path file = new Path(hhpath);
					FileSystem fs = file.getFileSystem(new Configuration());
					FSDataInputStream ds = fs.open(file);
					BufferedReader reader = new BufferedReader(new InputStreamReader(ds));
					
					String line;
					while((line = reader.readLine()) != null)
					{
						int index = line.indexOf("\t");
						if(index != -1)
						{
							String cidstr = line.substring(0, index);
							int cid = Integer.parseInt(cidstr);
							
							String values = line.substring(index+1);
							String fields[] = values.split(",");
							
							double ev = 0;
							for(int i=0; i<fields.length; i++)
							{
								ev = Double.parseDouble(fields[i]);
								hh[i][cid] = ev;
							}
						}
					}
					reader.close();
					ds.close();
					
					String ypath = context.getConfiguration().get("ypath");
					Path yfile = new Path(ypath);
					FileSystem hdfs = yfile.getFileSystem(new Configuration());
					yds = hdfs.open(yfile);
					readerinmap = new BufferedReader(new InputStreamReader(yds));
					
				} catch (Exception e) {

				}
			}
				
			public void map(LongWritable key, Text value, Context context)
					throws IOException, NumberFormatException, InterruptedException {

				String delim = new String();
				if (value.toString().contains("\t")) {
					delim = "\t";
				} else if (value.toString().contains("::")) {
					delim = "::";
				}
				String[] vals = value.toString().split(delim);
				
				int row = (int)key.get();
				
				double[] Wi = new double[k];
				for(int i=0; i<vals.length; i++)
				{
					Wi[i] = Double.parseDouble(vals[i]);
				}
				
				double[] tempresult = new double[k];
				
				double tempvalue = 0;
				for(int column=0; column<k; column++)
				{
					tempvalue = 0;
					for(int i=0; i<vals.length; i++)
					{
						tempvalue += Wi[i] * hh[i][column];
					}
					tempresult[column] = tempvalue;
				}
				
				double[] Y = new double[k];
				String line;
				if((line = readerinmap.readLine()) != null)
				{
					int index = line.indexOf("\t");
					if(index != -1)
					{
						String cidstr = line.substring(0, index);
						int cid = Integer.parseInt(cidstr);
						
						String values = line.substring(index+1);
						String fields[] = values.split(",");
						
						double ev = 0;
						for(int i=0; i<fields.length; i++)
						{
							ev = Double.parseDouble(fields[i]);
							Y[i] = ev;
						}
					}
				}
				
				double[] result = new double[k];
				String outputstr = "";
				for(int column=0; column<k; column++)
				{
					result[column] = Wi[column]*(Y[column]/tempresult[column]);
					
					if(column != k-1)
						outputstr += result[column] + ",";
					else
						outputstr += result[column];
				}
				
				context.write(new LongWritable(row), new Text(outputstr));

			}
				
			@Override
			protected void cleanup(Context context) throws IOException,
						InterruptedException {
				readerinmap.close();
				yds.close();
			}

		}
			
		//no need reducer in this job
		public static class wreduce3 extends
		Reducer<LongWritable, Text, LongWritable, Text> {
				
			private static double[][] ah; 
			int k = 10;
			int M = 100;
				
			@Override
			protected void setup(Context context) throws IOException,
					InterruptedException {
				k = Integer.parseInt(context.getConfiguration().get("k"));
				M = Integer.parseInt(context.getConfiguration().get("M"));
					
			}

			public void reduce(LongWritable key, Iterable<Text> values,
					Context context) throws IOException, InterruptedException {
					
			}
		}
		*/
	
		public static void main(String[] args) {
		// TODO Auto-generated method stub

		}

}
