package lossfunction;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;

import util.NonSplittableTextInputFormat;

public class LossBlockDifferWH {

	//compute W(T)*A
	public static class lossblockfremapdiff extends
	Mapper<LongWritable, Text, IntWritable, DoubleWritable> {
			
		private static double[][] w;
		private static double[][] h;
		
		private int k;
		private int M;
		private int N;
		private int RP; // rows are split in to RP parts
		private int CP; // columns are split in to CP parts
		private int itr; // iteration
		
		private int RperPart;
		private int CperPart;
		
		private boolean loadw = false;
		private boolean loadh = false;
		private int highrow = 0;
		private int highcolumn = 0;
		
		private double partialloss = 0;
		
		private int startrow = 0;
		private int startcolumn = 0;
		
		private boolean initialized = false;
		
		/*
		public int getRowPartNum(int i)
		{
			int num = 0;
			
			num = i/(M/RP);
			
			return num;
		}
		
		public int getColumnPartNum(int j)
		{
			int num = 0;
			
			num = j/(N/CP);
			
			return num;
		}
		*/
		
		@Override
		protected void setup(Context context) throws IOException,
				InterruptedException {
			
			k = Integer.parseInt(context.getConfiguration().get("k"));
			M = Integer.parseInt(context.getConfiguration().get("M"));
			N = Integer.parseInt(context.getConfiguration().get("N"));
			RP = Integer.parseInt(context.getConfiguration().get("RP"));
			CP = Integer.parseInt(context.getConfiguration().get("CP"));
			itr = Integer.parseInt(context.getConfiguration().get("itr"));
			
			partialloss = 0;
				
			if(M%RP == 0)
				RperPart = M/RP;
			else
				RperPart = M/RP + 1;
			
			if(N%CP == 0)
				CperPart = N/CP;
			else
				CperPart = N/CP + 1;
			
			w = new double[RperPart][k];
			h = new double[k][CperPart];
			
			for(int i=0; i<RperPart; i++)
				for(int j=0; j<k; j++)
				{
					w[i][j] = 0;
				}
			
			for(int i=0; i<k; i++)
				for(int j=0; j<CperPart; j++)
				{
					h[i][j] = 0;
				}
		}
			
		public void map(LongWritable key, Text value, Context context)
				throws IOException, NumberFormatException, InterruptedException {

//			String delim = new String();
//			if (value.toString().contains("\t")) {
//				delim = "\t";
//			} else if (value.toString().contains("::")) {
//				delim = "::";
//			}
//			String[] vals = value.toString().split(delim);
			
			int row = 0;
			int column = 0;
			double element = 0;
			
			String[] fields = value.toString().split(",");
			
			//if(fields.length != 3)
			//	System.out.println(":" + value.toString() + ":");
			
			row = Integer.parseInt(fields[0]);
			column = Integer.parseInt(fields[1]);
			element = Double.parseDouble(fields[2]);
			
			
			if(!initialized)
			{
				initialized = true;
				
				highrow = row/RperPart;
				highcolumn = column/CperPart;
				
				//System.out.println("column: " + column + ";   CperPart: " + CperPart);
				//System.out.println("in map, highcolumn:" + highcolumn);
				
				startrow = highrow * RperPart;
				startcolumn = highcolumn * CperPart;
			}
			
			
			if(!loadw)
			{
				loadw = true;
				
				try {
					String mdir = context.getConfiguration().get("middledir");
					
					String wpath = mdir + "/wmatrix/" + (itr-1) + "/w" + highrow;
					//generate final path using highrow
					Path file = new Path(wpath);
					FileSystem fs = FileSystem.get(new Configuration());
					
					int begin = itr - 1; //search forward
			    	while(!fs.exists(file))
			    	{
			    		begin--;
			    		file = new Path(mdir + "/wmatrix/" + begin + "/w" + highrow);
			    	}
			    	
			    	System.out.println("load file w: " + file.toString());
					
					FSDataInputStream ds = fs.open(file);
					BufferedReader reader = new BufferedReader(new InputStreamReader(ds));
					
					String line;
					int count = 0;
					while((line = reader.readLine()) != null)
					{
						String tmpfields[] = line.split(",");
						
						double ev = 0;
						for(int i=0; i<tmpfields.length; i++)
						{
							ev = Double.parseDouble(tmpfields[i]);
							w[count][i] = ev;
						}
						count++;
					}
					reader.close();
					ds.close();
				} catch (Exception e) {

				}
			}
			
			if(!loadh)
			{
				loadh = true;
				
				try {
					String mdir = context.getConfiguration().get("middledir");
					
					String hpath = mdir + "/hmatrix/" + itr + "/h" + highcolumn;
					//generate final path using highcolumn
					Path file = new Path(hpath);
					FileSystem fs = file.getFileSystem(new Configuration());
					FSDataInputStream ds = fs.open(file);
					BufferedReader reader = new BufferedReader(new InputStreamReader(ds));
					
					String line;
					int count = 0;
					while((line = reader.readLine()) != null)
					{
						fields = line.split(",");
						
						double ev = 0;
						for(int i=0; i<fields.length; i++)
						{
							ev = Double.parseDouble(fields[i]);
							h[i][count] = ev;
						}
						count++;
					}
					reader.close();
					ds.close();
				} catch (Exception e) {

				}
			}
			
			double tempproduct = 0;
			for(int j=0; j<k; j++)
			{
				tempproduct += w[row-startrow][j] * h[j][column-startcolumn];  
			}
			
			partialloss += (element - tempproduct) * (element - tempproduct);
		}
			
			@Override
			protected void cleanup(Context context) throws IOException,
					InterruptedException {
				
				context.write(new IntWritable(0), new DoubleWritable(partialloss));
			}

		}
		
		public static class lossblockfrereducediff extends
		Reducer<IntWritable, DoubleWritable, IntWritable, DoubleWritable> {

			public void reduce(IntWritable key, Iterable<DoubleWritable> values,
					Context context) throws IOException, InterruptedException {
		
				double totalloss = 0;
				for (DoubleWritable value : values) {
					totalloss += value.get();
				}
				context.write(new IntWritable(0), new DoubleWritable(totalloss));
			}
		}
		
		public static void computeloss(String input, String output, String middledir, int k, int itr, int m, int n, int rp, int cp) throws IOException,
		InterruptedException, ClassNotFoundException {
			//job1
			Configuration conf = new Configuration();
			conf.set("middledir", middledir);
			//conf.setBoolean("mapred.output.compression.codec", true);
			//conf.setClass("mapred.output.compression.codec", GzipCodec.class,CompressionCodec.class);
			//conf.setInt("mapred.inmem.merge.threshold",0);
			//conf.setFloat("mapred.job.reduce.input.buffer.percent", 1);
			conf.set("mapred.child.java.opts", "-Xmx400m");
			//conf.setInt("io.sort.mb", 200);
			conf.setInt("io.sort.factor", 100);
			//conf.setInt("mapred.reduce.parallel.copies",10);
			conf.setInt("io.file.buffer.size", 128);
			conf.set("k", k + "");
			conf.set("itr", itr + "");
			conf.set("M", m + "");
			conf.set("N", n + "");
			conf.set("RP", rp + "");
			conf.set("CP", cp + "");
			Job job = new Job(conf, "LossBlockfre");
			job.setJarByClass(LossBlockDifferWH.class);
			job.setNumReduceTasks(1);
			job.setJobName("LossBlockfre");
			job.setOutputKeyClass(IntWritable.class);
			job.setOutputValueClass(DoubleWritable.class);
			job.setMapperClass(lossblockfremapdiff.class);
			job.setReducerClass(lossblockfrereducediff.class);
			job.setInputFormatClass(NonSplittableTextInputFormat.class);
			job.setOutputFormatClass(TextOutputFormat.class);
			
			FileInputFormat.addInputPath(job, new Path(input));
			FileOutputFormat.setOutputPath(job, new Path(output));
			
			job.waitForCompletion(true);
		}
	
		public static void main(String[] args) throws IOException, InterruptedException, ClassNotFoundException {
		// TODO Auto-generated method stub

			if (args.length != 9) {
	            System.err.println("Usage: MainNMF <input> <output> <wh_location> <m> <n> <rp> <cp> <k> <itr> ");
	            System.exit(2);
	        }
			
			String input = args[0];
		    String output = args[1];
		    String middledir = args[2];
		    int m  = Integer.parseInt(args[3]);
		    int n  = Integer.parseInt(args[4]);
		    int rp  = Integer.parseInt(args[5]);
		    int cp  = Integer.parseInt(args[6]);
		    int k = Integer.parseInt(args[7]);
		    int itr = Integer.parseInt(args[8]);
		    
		    
		    //complete w blocks in the specified iteration
//		    FileSystem fs = FileSystem.get(new Configuration());
//		    String wpath = middledir + "/wmatrix/" + itr + "/w";
//		    for(int i=0; i<rp; i++)
//		    {
//		    	Path currentpath = new Path(wpath + i);
//		    	int begin = itr; //search forward
//		    	while(!fs.exists(currentpath))
//		    	{
//		    		begin--;
//		    		currentpath = new Path(middledir + "/wmatrix/" + begin + "/w" + i);
//		    	}
//		    	String src = middledir + "/wmatrix/" + begin + "/w" + i;
//		    	String dst = middledir + "/wmatrix/" + itr + "/";
//		    	Process p = Runtime.getRuntime().exec("/home/jyin/hadoop-1.0.3/bin/hadoop dfs -cp " + src + " " + dst);
//		    	p.waitFor();
//		    }
		    
		    for(int i=1; i<=itr; i++)
		    {
		    	String newoutput = output + "/" + i;
		    	computeloss(input, newoutput, middledir, k, i, m, n, rp, cp);
		    }
		}

}
